import { Injectable } from '@angular/core'
import { EventService } from './event.service'
import { globals } from '../_globals'
import { SasService } from './sas.service'
import { NavigationEnd, Router } from '@angular/router'
import { BehaviorSubject } from 'rxjs'
import { LoggerService } from './logger.service'
import { EnvironmentInfo } from '../system/models/environment-info.model'
import { LicenceService } from './licence.service'

@Injectable()
export class AppService {
  public syssite = new BehaviorSubject<string[] | null>(null)
  private environmentInfo: EnvironmentInfo | null = null

  constructor(
    private licenceService: LicenceService,
    private eventService: EventService,
    private sasService: SasService,
    private loggerService: LoggerService,
    private router: Router
  ) {
    this.subscribe()

    this.router.events.subscribe((val) => {
      if (val instanceof NavigationEnd) {
        if (val.url === '/deploy') {
          this.eventService.startupDataLoaded()
        }
      }
    })
  }

  sasServiceInit() {
    this.sasService.sasServiceInit() //Code from constructor is put in this function inside SasService, because we first want to subscribe to emitters here before executing SasService init function
  }

  subscribe() {
    this.sasService.loadStartupServiceEmitter.subscribe(() => {
      this.startUpData()
    })

    this.sasService.requestSiteIdEmitter.subscribe((sysite: string) => {
      this.patchSyssite(sysite)
    })
  }

  public getEnvironmentInfo() {
    return this.environmentInfo
  }

  public patchSyssite(syssite: string) {
    let currentSyssite = this.syssite.getValue()

    if (currentSyssite) {
      if (!currentSyssite.includes(syssite)) {
        currentSyssite.push(syssite)
        this.syssite.next(currentSyssite)
      }
    }
  }

  public async startUpData() {
    let startupServiceError = false

    await this.sasService
      .request('public/startupservice', null)
      .then(async (res: any) => {
        this.syssite.next([res.SYSSITE])

        let missingProps: string[] = []

        if (!res.globvars || (res.globvars && !res.globvars[0]))
          missingProps.push('Globvars')
        if (!res.sasdatasets) missingProps.push('Sasdatasets')
        if (!res.saslibs) missingProps.push('Saslibs')

        if (missingProps.length > 0) {
          startupServiceError = true
          this.eventService.showInfoModal(
            'Error',
            `${missingProps.join(', ')} are not present in the startupservice`
          )
          this.licenceService.isAppActivated.next(false)

          return
        }

        this.environmentInfo = {
          SYSSITE: res.SYSSITE,
          SYSSCPL: res.SYSSCPL,
          SYSTCPIPHOSTNAME: res.SYSTCPIPHOSTNAME,
          SYSVLONG: res.SYSVLONG,
          MEMSIZE: res.MEMSIZE,
          SYSPROCESSMODE: res.SYSPROCESSMODE,
          SYSHOSTNAME: res.SYSHOSTNAME,
          SYSHOSTINFOLONG: res.SYSHOSTINFOLONG,
          SYSENCODING: res.SYSENCODING,
          AUTOEXEC: res.AUTOEXEC,
          ISADMIN: res.globvars[0].ISADMIN,
          DC_ADMIN_GROUP: res.globvars[0].DC_ADMIN_GROUP
        }

        let libs = res.sasdatasets
        let libGroup: any = {}
        let libsAndTables
        let libraries

        for (let lib of libs) {
          if (!libGroup[lib.LIBREF]) {
            libGroup[lib.LIBREF] = []
          }

          libGroup[lib.LIBREF].push(lib.DSN)
        }

        // Sidebar data structure
        let treeNodeTemp = JSON.parse(JSON.stringify(libGroup))
        let treeNode = []

        for (let library of Object.keys(treeNodeTemp)) {
          treeNode.push({
            LIBRARYREF: library,
            tables: treeNodeTemp[library]
          })
        }

        let treeNodeLibraries = treeNode
        // Sidebar data structure end

        libsAndTables = libGroup
        libraries = Object.keys(libGroup)

        if (libsAndTables !== undefined) {
          globals.editor.libsAndTables = libsAndTables
        }

        globals.editor.treeNodeLibraries = treeNodeLibraries
        globals.editor.libraries = libraries
        globals.editor.startupSet = true

        await this.licenceService.activation(res)
      })
      .catch((err: any) => {
        startupServiceError = true
        this.eventService.showInfoModal(
          'Error',
          'There is an issue with startupservice response'
        )
        this.licenceService.isAppActivated.next(false)
      })

    this.loggerService.log(
      'Activated:',
      this.licenceService.isAppActivated.value
    )

    if (!startupServiceError) {
      this.eventService.startupDataLoaded()

      if (!this.licenceService.isAppActivated.value) {
        // If route is already on `licensing` page then we won't re route
        if (
          !this.router.url.includes('deploy') &&
          !this.router.url.includes('licensing')
        ) {
          this.router.navigateByUrl('/licensing/key?error=missing&force=true')
        }
      } else {
        if (this.router.url.includes('licensing'))
          this.router.navigateByUrl('/home')
      }
    }
  }
}
